int dspset=1;
unsigned short *ramw;
unsigned short dspram[0x1000];

int dsprun=0;
unsigned long dsppc;
unsigned long dspflags=0;
int dspendian;

unsigned readdsp(unsigned addr)
{
        switch (addr)
        {
                case 0xF1A100: /*D_FLAGS high*/
                return dspflags>>16;
                case 0xF1A102: /*D_FLAGS low*/
                return dspflags;

                default:
                printf("Bad DSP read %08X\n",addr);
                exit(-1);
        }
}

void writedsp(unsigned addr, unsigned val)
{
        switch (addr)
        {
                case 0xF1A100: /*D_FLAGS high*/
                dspflags=(dspflags&0xFFFF)|(val<<16);
                break;
                case 0xF1A102: /*D_FLAGS low*/
                dspflags=(dspflags&0xFFFF0000)|val;
                dspset=(dspflags&0x4000)?1:0;
                break;

                case 0xF1A10C: case 0xF1A10E: /*Endian control*/
                dspendian=val&7;
                printf("dsp endian %i\n",val&7);
                break;

                case 0xF1A110: /*D_PC high*/
                dsppc=(dsppc&0xFFFF)|(val<<16);
                break;

                case 0xF1A112: /*D_PC low*/
                dsppc=(dsppc&0xFFFF0000)|val;
                printf("D_PC %06X\n",dsppc);
                break;

                case 0xF1A114: /*D_CTRL high*/
                break;

                case 0xF1A116: /*D_CTRL low*/
                dsprun=val&1;
                break;

                default:
                printf("Bad DSP write %08X %08X\n",addr,val);
                exit(-1);
        }
}

/*DSP emulation*/
unsigned long dspregs[2][32];

void dumpdspregs()
{
        int c;
        printf("PC=%08X FLAGS=%08X\n",dsppc,dspflags);
        for (c=0;c<8;c++)
            printf("R%02i=%08X R%02i=%08X R%02i=%08X R%02i=%08X\n",c<<2,dspregs[1][c<<2],(c<<2)+1,dspregs[1][(c<<2)+1],(c<<2)+2,dspregs[1][(c<<2)+2],(c<<2)+3,dspregs[1][(c<<2)+3]);
}

inline unsigned short dquickread(unsigned long addr)
{
        if (addr<0x200000) return ramw[addr>>1];
        if ((addr>=0xF1B000) && (addr<0xF1D000)) return dspram[((addr-0x1000)&0x1FFF)>>1];
        printf("Bad dsp read %08X\n",addr);
        exit(-1);
}

inline unsigned short dread(unsigned long addr)
{
        if ((addr>=0xF1B000) && (addr<0xF1D000)) return dspram[((addr-0x1000)&0x1FFF)>>1];
        if (addr<0x200000) return ramw[addr>>1];
        if ((addr>=0xF1A100) && (addr<0xF1A120)) return readdsp(addr);
        printf("Bad dsp read %08X\n",addr);
        exit(-1);
}

inline void dwrite(unsigned long addr, unsigned short val)
{
        if ((addr>=0xF1A100) && (addr<0xF1A120))
        {
                writedsp(addr,val);
                return;
        }
        printf("Bad dsp write %08X %04X\n",addr,val);
        dumpdspregs();
        exit(-1);
}

void dspb0rk()
{
        printf("dsp B0rked\n");
        dumpdspregs();
        exit(-1);
}

#define RN (opcode&31)
#define RM ((opcode>>5)&31)

#define C0 (opcode&0x20)
#define C1 (opcode&0x40)
#define C2 (opcode&0x80)
#define C3 (opcode&0x100)
#define C4 (opcode&0x200)

#define ZFLAG (dspflags&1)
#define CFLAG (dspflags&2)
#define NFLAG (dspflags&4)

inline void setdspzn(unsigned long v)
{
        dspflags&=~5;
        if (!v)                dspflags|=ZFLAG;
        else if (v&0x80000000) dspflags|=NFLAG;
}

void rundsp(int cycles)
{
        unsigned short opcode;
        while (cycles>0)
        {
                opcode=dquickread(dsppc);
                dsppc+=2;
                switch (opcode>>10)
                {
                        case 14: /*BSET #,Rn*/
                        dspregs[dspset][RN]|=(1<<RM);
                        setdspzn(dspregs[dspset][RN]);
                        cycles-=3;
                        break;

                        case 35: /*MOVEQ*/
                        dspregs[dspset][RN]=RM;
                        cycles-=3;
                        break;

                        case 36: /*MOVETA*/
                        dspregs[dspset^1][RN]=dspregs[dspset][RM];
                        cycles-=3;
                        break;

                        case 38: /*MOVEI*/
                        dspregs[dspset][RN]=dquickread(dsppc)|(dquickread(dsppc+2)<<16);
                        dsppc+=4;
                        cycles-=3;
                        break;

                        case 41: /*LOAD (Rm),Rn*/
                        if (dspregs[dspset][RM]&3) dspb0rk();
                        dspregs[dspset][RN]=dread(dspregs[dspset][RM]+2)|(dread(dspregs[dspset][RM])<<16);
                        cycles-=3;
                        break;

                        case 47: /*STORE Rn,(Rm)*/
                        if (dspregs[dspset][RM]&3) dspb0rk();
                        dwrite(dspregs[dspset][RM]+2,dspregs[dspset][RN]);
                        dwrite(dspregs[dspset][RM],dspregs[dspset][RN]>>16);
                        cycles-=3;
                        break;

                        case 52: /*JUMP cc,(Rn)*/
                        if ((!C0 || !ZFLAG) &&
                            (!C1 || ZFLAG) &&
                            (!C2 || ((C4)?!NFLAG:!CFLAG)) &&
                            (!C3 || ((C4)?NFLAG:CFLAG)))
                        {
                                dsppc=dspregs[dspset][RN];
                                printf("C were %i\n",(opcode>>5)&31);
                        }
                        cycles-=3;
                        break;

                        case 57: /*NOP*/
                        cycles-=3;
                        break;

                        default:
                        printf("Bad dsp opcode %i %04X\n",opcode>>10,opcode);
                        dumpdspregs();
                        exit(-1);
                }
                if (!dsprun) return;
        }
}
